/*********************************************************/
/* mipmap: experimental environment for					 */
/*         texturing, mipmaps and filtering				 */
/*********************************************************/
/* Daniel Bielser, ETH Zurich - january  2001			 */
/*  adapted by Thomas Oskam, ETH Zurich - nov 2005       */
/*********************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>

#include "GL/glut.h"


/* default window size */
#define WINDOW_WIDTH 1000
#define WINDOW_HEIGHT 800

/* parameter for anisotropic filtering */
#define GL_TEXTURE_MAX_ANISOTROPY_EXT 0x84FE
#define GL_MAX_TEXTURE_MAX_ANISOTROPY_EXT 0x84FF

bool anisotropic_supported = false;
bool use_anisotropic = false;
float maxAnisotropy = 0;

/* number of trajectory subdivisions for... */
#define GEOMETRY_SUBDIVS    15.0  /* ...the geometrical representation */
#define TRAJECTORY_SUBDIVS  60.0  /* ...the camera path */

#define NBR_OF_ENTRIES 43	    /* number of control points descriping the trajectory */
#define TRACK_BREADTH  1.0      /* half breadth of the track */

#define CAMERA_HIGHT 1.0	    /* camera hight above the track */

#define TEXTURE_SIZE 512	    /* size of the base texture */
#define NBR_OF_MIPMAP_LEVELS 10 /* number of mipmap levels */

/* mipmap-textures */
#define TX_COLORED       0  /* separate color for each mip map */
#define TX_CHECKER_BOARD 1  /* black and white checkerboard-texture */
#define TX_FILE 		 2  /* texture loaded from a file */

/* mipmap generation filters */
#define TX_NEAREST_NEIGHBOR_FILTER 0
#define TX_BOX_FILTER              1
#define TX_BILINEAR_FILTER         2

#define ERR_BAD_FILE_READ -1 /* error during parsing of texture file */

/*********************/
/* mip-map texturing */
/*********************/

struct Pixel  /* stores the color values of a single pixel */
{
  GLubyte red;
  GLubyte green;
  GLubyte blue;
};

Pixel mipMapTexture[NBR_OF_MIPMAP_LEVELS][TEXTURE_SIZE][TEXTURE_SIZE]; /* the texture pyramide for generating the mip-maps */

char *texture_file_name = "Textures/cgl_logo512.ppm";
int texture_type = TX_CHECKER_BOARD;   /* type of applied texture (TX_COLORED, TX_LINE_PATTERN, or TX_FILE) */
int txMipMapGenFilter = TX_BOX_FILTER; /* type of texture generation filter (TX_NEAREST_NEIGHBOR_FILTER, TX_BOX_FILTER, or TX_LINEAR_FILTER) */
int txMipMapFilter = GL_NEAREST;       /* type of texture filter (GL_NEAREST, GL_NEAREST_MIPMAP_NEAREST, GL_LINEAR_MIPMAP_NEAREST, 
																  or GL_LINEAR_MIPMAP_LINEAR) */
																  
/**************/
/* trajectory */
/**************/

GLboolean stop = GL_FALSE; /* current motion state */ 

float subdivs;           /* number of subdivisions of the trajecory */

float pos[3];            /* current position on the track */ 
float prev_pos[3];       /* previous position on the track */
float next_pos[3];       /* next position on the track */
float tang[3];           /* current tangent vector */
float prev_tang[3];      /* next tangent vector */
float inter_pos[3];      /* interpolated position */
float next_inter_pos[3]; /* following interpolated position */

/* counters for the qubic interpolation of the trajecory */
int segment_nbr;         /* counts the number of segment of the bezier curve */     
float subdiv_nbr;	     /* counts the subdivision step inside one curve segment */

/* bezier points of the trajectory */
float pos_array[NBR_OF_ENTRIES][3] = {{0,   0,   0},  
									  {0,   0,  -5},
									  {0,   0, -10},
									  {0,   0, -20},
									  {0,  10, -30},
									  {0,  25, -45},
									  {0,  40, -60},
									  {0,  55, -75},
									  {0,  70, -90},
									  {0,  80,-100},
									  {0,  80,-110},
									  {0,  80,-120},
									  {10, 80,-120},
									  {20, 80,-120},
									  {30, 80,-120},
									  {40, 80,-120},
									  {41, 70,-120},
									  {43, 50,-120},
									  {45, 30,-120},
									  {46, 20,-120},
									  {50, 20,-120},
									  {60, 20,-120},
									  {70, 20,-120},
									  {74, 20,-120},
									  {75, 30,-120},
									  {77, 45,-120},
									  {79, 60,-120},
									  {80, 70,-120},
									  {90, 70,-120},									  
									  {100,70,-120},
									  {100,65,-110},
									  {100,60,-100},
									  {80, 50,-100},
									  {60, 40,-100},
									  {60, 30,-120},
									  {60, 20,-140},
									  {80, 10,-140},
									  {100,0, -140},
									  {100,0, -120},
									  {100,0, -100},									
									  {50, 0, -50},
									  {0,  0, 10},
									  {0,  0, 0}};

GLfloat camera_speed = 1;
									  

/****************************************************************************************/
/* getMipMapPixel                                                                       */
/****************************************************************************************/
/* returns a pointer to a mipmap-pixel stored in the mipmap-pyramid, which is          .*/
/* referenced by its coordinates and its mipmap level                                   */
/* if the coordinates are outside of the specified mipmap, wrapping (repeating) is      */
/* performed                                                                            */     
/****************************************************************************************/
/* --> x,y     the coordinates of the pixel inside the mipmap-texture of the specified  */
/*             level different levels have different array boundaries					*/												
/* --> level   the level of the mipmap-texture inside the mipmap-pyramide 			    */
/*             (between 0 and NBR_OF_MIPMAP_LEVEL)										*/
/* <-- pointer to the requested pixel													*/
/****************************************************************************************/ 
Pixel* getMipMapPixel(GLint x, GLint y, GLint level)
{
  GLint level_width; /* the dimension of the given texture level */
  GLint r, s;        /* the texture indices after border wrapping */

  /* calc. the texture dimension of level */
  level_width = (TEXTURE_SIZE>>level);

  // handle negative values
  while(x < 0)
	x = level_width + x;
	
  while(y < 0)
	y = level_width + y;	

  /* border wrapping (necessary for some filters with large support) */
  r = x % level_width;
  s = y % level_width;

  /* return the texture values of this level */
  return &mipMapTexture[level][s][r];
}


/****************************************************************************************/
/* nearestNeighborFilter                                                                */
/****************************************************************************************/
/* Constructs the texture of the given mipmap-level from the data of the previous level.*/
/* A pixel is set always to the value of the nearest pixel in the texture of the        */
/* previous mipmap level. 																*/
/****************************************************************************************/
/* --> level   the mipmap level in the mipmap-pyramid 								    */
/*             (between 1 and NBR_OF_MIPMAP_LEVEL)										*/
/****************************************************************************************/ 
void nearestNeighborFilter(GLint level)
{
  Pixel *p;   /* pointer to the pixel in the mipmap texture a level higher */
  Pixel *p1;  /* pointer to the pixel in the current mipmap texture */
  GLint size; /* dimension of the current mipmap texture */
  GLint x, y; /* counters */
  
  GLubyte textureArray[TEXTURE_SIZE*TEXTURE_SIZE][3]; /* array for OpenGL texture definition */

  /* calculate the dimension of current mipmap level */
  size = (TEXTURE_SIZE>>level);
  
  /* for each mipmap pixel of current mipmap level .. */
  for (y=0; y<size; y++)
  {
	for (x=0; x<size; x++)
	{
	  /* get the pointer to the pixel value of the mipmap texture a level higher */
	  p1 = getMipMapPixel(x*2,   y*2, level-1);

	  /* get the pointer to the pixel value of the current mipmap texture */
	  p = getMipMapPixel(x, y, level);

	  /* assign the intensities to the current pixel of the mipmap-pyramide */
	  p->red   =  p1->red;
	  p->green =  p1->green;
	  p->blue  =  p1->blue;

	  /* and store them into the array for the OpenGL texture definition */
	  textureArray[y*size+x][0] = p->red;
	  textureArray[y*size+x][1] = p->green;
	  textureArray[y*size+x][2] = p->blue;
	}
  }
  
  /* define the mip-map-texture of the given level in OpenGL */ 
  glTexImage2D(GL_TEXTURE_2D, level, 3, size, size, 0,
			   GL_RGB, GL_UNSIGNED_BYTE, textureArray);
}


/****************************************************************************************/
/* boxFilter                                                                            */
/****************************************************************************************/
/* Constructs the texture of the given mipmap-level from the data of the previous level.*/
/* A pixel is set to the averaged red, green, and blue intensities of the 4 neighboring */
/* pixels in the texture of the previous mipmap level. 							        */
/****************************************************************************************/
/* --> level   the mipmap level in the mipmap-pyramid									*/
/*             (between 1 and NBR_OF_MIPMAP_LEVEL)										*/
/****************************************************************************************/ 
void boxFilter(GLint level)
{
  Pixel *p;                 /* pointer to the pixel in the current mipmap texture */
  Pixel *p1, *p2, *p3, *p4; /* pointers to the pixel in the mipmap texture a level higher */
  GLint size;               /* dimension of the current mipmap texture */
  GLint x, y;               /* counters */
  
  GLubyte textureArray[TEXTURE_SIZE*TEXTURE_SIZE][3]; /* array for OpenGL texture definition */




  /* insert here your solution of question 2a) */



 
}

/****************************************************************************************/
/* bilinearFilter                                                                       */
/****************************************************************************************/
/* Constructs the texture of the given mipmap-level from the data of the previous level.*/
/* A pixel is set to the weighted average of the red, green, and blue intensities of    */
/* the pixels in the texture of the previous mipmap level. The weights are defined by a */
/* filter-matrix.																	    */
/****************************************************************************************/
/* --> level   the mipmap level in the mipmap-pyramid									*/
/*             (between 1 and NBR_OF_MIPMAP_LEVEL)										*/
/****************************************************************************************/ 
void bilinearFilter(GLint level)
{
  Pixel *p[5][5];/* pointers to the pixel in the mipmap texture a level higher */
  Pixel *r;      /* pointer to the pixel in the current mipmap texture */

  GLfloat red, green, blue;

  GLint size;   /* dimension of the current mipmap texture */
  GLint x, y;   /* counters for reverencing the pixels of the current mipmap*/
  GLint i, j;   /* counters for reverencing the filter matrix entries */

  GLubyte textureArray[TEXTURE_SIZE*TEXTURE_SIZE][3]; /* array for OpenGL texture definition */

  /* define separable bilinear filter */
  GLfloat lin5[5][5]={
    {1., 2., 3., 2., 1.},
    {2., 4., 6., 4., 2.},
    {3., 6., 9., 6., 3.},
    {2., 4., 6., 4., 2.},
    {1., 2., 3., 2., 1.}};

  /* normalize bilinear filter */





  /* insert here your solution of question 2b) */




}


/****************************************************************************************/
/* coloredMipMap                                                                        */
/****************************************************************************************/
/* constructs and defines a mipmap-texture with a color that depends on the mip-map     */
/* level																				*/
/****************************************************************************************/
/* --> level   the mipmap level in the mipmap-pyramid									*/
/****************************************************************************************/
void coloredMipMap(int level)
{
  int i, j;        /* counters */
  Pixel color;     /* color value (RGB) */

  GLint size;      /* texture dimension of the mipmap level */

  GLubyte textureArray[TEXTURE_SIZE*TEXTURE_SIZE][3]; /* linear array for OpenGL texture definition */


  /* insert here your solution of question 1a) */
 ///////////////////////////////////////////////

  /* calculate texture dimension (size) of the current mipmap level */


  /* determine a color value (color) for the current mipmap level */
 

  /* define the texture (textureArray) with the current color value */

  
  /* define the mip-map-texture of the given level in OpenGL (textureArray) */ 
  
}	


/* generateMipMap:
 * <-- pointer to actual object                              
 * generates all mipmaps for given object                    
 */
/****************************************************************************************/
/* generateMipMaps                                                              */
/****************************************************************************************/
/* constructs a checkerboard texture and defines it as the first level of 				*/
/* the mipmap-texture 																	*/
/****************************************************************************************/
void generateMipMaps()
{
  GLint level;
	
  /* while not last mipmap level proceeded */
  for(level = 1; level < NBR_OF_MIPMAP_LEVELS; level++)
  {
	if(texture_type == TX_COLORED)	
	{
	  coloredMipMap(level);
	}	  
	else
	{
	  if (txMipMapGenFilter == TX_NEAREST_NEIGHBOR_FILTER)
	  {
		nearestNeighborFilter(level);
	  }		
	  else if(txMipMapGenFilter == TX_BOX_FILTER)
	  {
		boxFilter(level);
	  }
	  else if (txMipMapGenFilter == TX_BILINEAR_FILTER)
	  {
		bilinearFilter(level);
	  }		
	}
  }
}


/****************************************************************************************/
/* makeCheckerBoardPattern                                                              */
/****************************************************************************************/
/* constructs a checkerboard texture and defines it as the first level of 				*/
/* the mipmap-texture 																	*/
/****************************************************************************************/
void makeCheckerBoardPattern(void)
{
  int i, j; /* counters */
  int c;    /* color value */
  
  GLubyte textureArray[TEXTURE_SIZE*TEXTURE_SIZE][3]; /* array for gl texture definition */

  /* traverse all texture pixels */
  for (i = 0; i < TEXTURE_SIZE; i++) 
  {
	for (j = 0; j < TEXTURE_SIZE; j++) 
	{
	  /* check if it should be a black or a white pixel (the diameter of a square is 82) */
	  if((((i/82)&0x1)==0)^(((j/82)&0x1)==0))
	  {
		c = 0;
	  }
	  else
	  {
		c = 255;
	  }
	  
	  /* define the color value of the texture pixel (i,j) of the first level of the mipmap texture  */
	  mipMapTexture[0][i][j].red = c;
	  mipMapTexture[0][i][j].green = c;
	  mipMapTexture[0][i][j].blue = c;
	  
	  /* define the color value of the texture pixel (i,j) for the OpenGL texture definition */
	  textureArray[i*TEXTURE_SIZE+j][0] = c;
	  textureArray[i*TEXTURE_SIZE+j][1] = c;
	  textureArray[i*TEXTURE_SIZE+j][2] = c;
	}
  }
  
  /* gl basis texture definition */
  glTexImage2D(GL_TEXTURE_2D, 0, 3, TEXTURE_SIZE, TEXTURE_SIZE, 0,
			   GL_RGB, GL_UNSIGNED_BYTE, textureArray);
}	


/****************************************************************************************/
/* readTexmap                                                                  			*/
/****************************************************************************************/
/* loads the texture as an irix rgb file from disk	      								*/
/****************************************************************************************/
/*  --> acFile     the filename														    */
/*  <-- pointer to the texture map														*/
/****************************************************************************************/
int readTexmap(char *acFile, GLubyte TexMap[TEXTURE_SIZE][TEXTURE_SIZE][3])
{
  FILE* in;
  int i, j, xsize, ysize;
  unsigned char ch, buf[200];
  
  in = fopen(acFile, "rb");
  if( !in ){
    printf("ERROR: dwsa_read_texmap: bad fileopen\n");
    glDisable(GL_TEXTURE_2D);
    return ERR_BAD_FILE_READ;
  }
  
  /*read header*/
  fread(buf, 2, 1, in); /*magic num*/
  if( (buf[0] != 'P') || (buf[1] != '6') ){
    printf("ERROR: dwsa_read_texmap: bad file magicnum\n");
    glDisable(GL_TEXTURE_2D);
    return ERR_BAD_FILE_READ;
  }
  /*skip whitespace*/
  fread(&buf[0], 1, 1, in);
  do
  {
    ch = buf[0];
    while( ch != '\n' ) fread(&ch, 1, 1, in);
	  fread(&buf[0], 1, 1, in);
  }while( (buf[0] < '0') || (buf[0] > '9') );
  /*get width*/
  fread(&ch, 1, 1, in);
  for(i = 1; (ch >= '0') && (ch <= '9'); i++)
  {
    buf[i] = ch;
    fread(&ch, 1, 1, in);
  }
  buf[i] = '\0';
  xsize = atoi((char *)buf);
  /*the space has been read, can continue straight*/
  fread(&ch, 1, 1, in);
  for(i = 0; (ch >= '0') && (ch <= '9'); i++)
  {
    buf[i] = ch;
    fread(&ch, 1, 1, in);
  }
  buf[i] = '\0';
  ysize = atoi((char *)buf);
  /*the newline has been read, can continue straight*/
  fread(&ch, 1, 1, in);
  for(i = 0; (ch >= '0') && (ch <= '9'); i++)
  {
    buf[i] = ch;
    fread(&ch, 1, 1, in);
  }
  buf[i] = '\0';
  
  if( (xsize != TEXTURE_SIZE) || (ysize != TEXTURE_SIZE) )
  {
    printf("ERROR: dwsa_read_texmap: bad file dims\n");
    return ERR_BAD_FILE_READ;
  }
   
  /*the newline is read, continue straight*/
  for(i = 0; i < xsize; i++)
  {
	for(j = 0; j < ysize; j++)
	{
	  fread(&ch, 1, 1, in);
	  TexMap[xsize-1-i][j][0] = ch;
	  fread(&ch, 1, 1, in);
	  TexMap[xsize-1-i][j][1] = ch;
	  fread(&ch, 1, 1, in);
	  TexMap[xsize-1-i][j][2] = ch;
	}
  }
  fclose(in);
  return 1;
}

/****************************************************************************************/
/* readTextureFile                                                                      */
/****************************************************************************************/
/* reads the texture file with the file name stored in the global variable              */
/* >texture_file_name<  and defines it as the first level of the mipmap-texture 		*/
/****************************************************************************************/
void readTextureFile(void)
{
  int i, j; /* counters */
  GLubyte textureArray[TEXTURE_SIZE][TEXTURE_SIZE][3]; /* array for gl texture definition */

  /* read the texture with the filename texture_file_name and give back a pointer to the first element (textureArray) */
  if(readTexmap(texture_file_name, textureArray))
  {
	/* store the texture in the first level of the mipmap pyramid */
	for (i = 0; i < TEXTURE_SIZE; i++) 
	{
	  for (j = 0; j < TEXTURE_SIZE; j++) 
	  {
		mipMapTexture[0][i][j].red =   textureArray[i][j][0];
		mipMapTexture[0][i][j].green = textureArray[i][j][1];
		mipMapTexture[0][i][j].blue =  textureArray[i][j][2];
	  } 
	}

	/* define this texture  as current OpenGL texture */
	glTexImage2D(GL_TEXTURE_2D, 0, 3, TEXTURE_SIZE, TEXTURE_SIZE, 0,
				 GL_RGB, GL_UNSIGNED_BYTE, textureArray);
  }
}





/****************************************************************************************/
/* initTexturing                                                                        */
/****************************************************************************************/
/* initializes the (mipmap-) textures and sets the appropriate OpenGL parameters     	*/
/****************************************************************************************/
void initTexturing(void)
{    
  glEnable(GL_DEPTH_TEST);
  
  /* sets the mode of pixel alignment in memory */  
  glPixelStorei(GL_UNPACK_ALIGNMENT,1);
    
  /* generate the specific basis texture depending on texture_type */
  switch(texture_type)
  {
	case TX_COLORED:	
	{
	  coloredMipMap(0);
	}
	break;
	case TX_FILE:
	{
	  readTextureFile();
	}
	break;
	case TX_CHECKER_BOARD:
	{	
	  makeCheckerBoardPattern();
	}
	break;
  }

  /* check anisotropy factor */
  if (anisotropic_supported) glGetFloatv(GL_MAX_TEXTURE_MAX_ANISOTROPY_EXT, &maxAnisotropy);
  		
  /* generate mip maps */
  generateMipMaps();

  /* define the filter operation for the display of textures very close to the camera (magnification filter) */
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);


  /* define the type of mipmap-filtering given in the variable txMipMapFilter with glTexParameteri(...) */


  /* enable texturing with glEnable(...) */


  /* aufgabe 3 b) */
  if (use_anisotropic) {
		// ............... //
  }

}




/****************************************************************************************/
/* initTrajectoryValue                                                                  */
/****************************************************************************************/
/* initializes the trajectory function used either for the object generation or      	*/
/* as camera path																		*/
/****************************************************************************************/
/*  --> new_subdivs   the number of subdivisions inbetween each pair of bezier points   */
/****************************************************************************************/
void initTrajectoryValue(int new_subdivs)
{
  /* set the resolution of the trajectory */
  subdivs = new_subdivs;
  
  /* init the interpolation counters */
  segment_nbr = 0;
  subdiv_nbr = 0;
  
  /* read the first positions from the trajectory array */
  prev_pos[0] = pos_array[0][0];
  prev_pos[1] = pos_array[0][1];
  prev_pos[2] = pos_array[0][2];

  pos[0] = pos_array[1][0];
  pos[1] = pos_array[1][1];
  pos[2] = pos_array[1][2];

  next_pos[0] = pos_array[2][0];
  next_pos[1] = pos_array[2][1];
  next_pos[2] = pos_array[2][2];
  
  /* init the tangent vector */
  prev_tang[0] = tang[0] = 0.0;
  prev_tang[1] = tang[1] = 0.0;
  prev_tang[2] = tang[2] = -1.0;

  /* set the interpolated value to the first position (inter_pos is initialized later) */
  next_inter_pos[0] = prev_pos[0];
  next_inter_pos[1] = prev_pos[1];
  next_inter_pos[2] = prev_pos[2];
}


/****************************************************************************************/
/* incrementTrajectoryValue                                                                  */
/****************************************************************************************/
/* increments the interpolated position on the trajectory defined by the control points */
/* (pos_array) by one interval            	                                            */
/****************************************************************************************/
/*  <--  return value is TRUE as long as the trajectory value is valid                  */
/****************************************************************************************/
int incrementTrajectoryValue()
{
  float inter_pos1[3]; /* results of the first linear interpolation step */
  float inter_pos2[3]; /* results of the second linear interpolation step */
  float tang_length;   /* length of the tangent vector */
  
  /* check if we are still in the same Bezier-segment */
  if(subdiv_nbr < subdivs)
  { /* yes, we are still inside the same segment */
	/* increment the interpolation counter */
	subdiv_nbr++;
  }
  else
  {	/* no, we entered a new segment */
	/* check if we have reached the end of the control-point array */
	if(segment_nbr < NBR_OF_ENTRIES-3)
	{ /* the end is not yet reached */
	  /* reinit the interpolation counters */
	  subdiv_nbr = 1;
	  segment_nbr += 2;

	  // fetch the new values
	  prev_pos[0] = pos_array[segment_nbr][0];
	  prev_pos[1] = pos_array[segment_nbr][1];
	  prev_pos[2] = pos_array[segment_nbr][2];

	  pos[0] = pos_array[segment_nbr+1][0];
	  pos[1] = pos_array[segment_nbr+1][1];
	  pos[2] = pos_array[segment_nbr+1][2];  

	  next_pos[0] = pos_array[segment_nbr+2][0];
	  next_pos[1] = pos_array[segment_nbr+2][1];
	  next_pos[2] = pos_array[segment_nbr+2][2];

	}
	else
	{ /* the end is reached */
      /* no valid position value left */
	  return GL_FALSE;
	}

  }  

  /* store the old interpolation values */
  inter_pos[0] = next_inter_pos[0];
  inter_pos[1] = next_inter_pos[1];
  inter_pos[2] = next_inter_pos[2];

  /* and calculate the new interpolation values with the casteljau algorithm */

  /* firstlinear interpolation -> interpos1, interpos2 */
  inter_pos1[0] = (subdivs-subdiv_nbr)/subdivs*prev_pos[0] + subdiv_nbr/subdivs*pos[0];
  inter_pos1[1] = (subdivs-subdiv_nbr)/subdivs*prev_pos[1] + subdiv_nbr/subdivs*pos[1];
  inter_pos1[2] = (subdivs-subdiv_nbr)/subdivs*prev_pos[2] + subdiv_nbr/subdivs*pos[2];  

  inter_pos2[0] = (subdivs-subdiv_nbr)/subdivs*pos[0] + subdiv_nbr/subdivs*next_pos[0];
  inter_pos2[1] = (subdivs-subdiv_nbr)/subdivs*pos[1] + subdiv_nbr/subdivs*next_pos[1];
  inter_pos2[2] = (subdivs-subdiv_nbr)/subdivs*pos[2] + subdiv_nbr/subdivs*next_pos[2];  

  /* second linear interpolation: interpos1, interpos2 -> next_inter_pos */
  next_inter_pos[0] = (subdivs-subdiv_nbr)/subdivs*inter_pos1[0] + subdiv_nbr/subdivs*inter_pos2[0];
  next_inter_pos[1] = (subdivs-subdiv_nbr)/subdivs*inter_pos1[1] + subdiv_nbr/subdivs*inter_pos2[1];
  next_inter_pos[2] = (subdivs-subdiv_nbr)/subdivs*inter_pos1[2] + subdiv_nbr/subdivs*inter_pos2[2];  

  /* store the old tangent vecotr*/
  prev_tang[0] = tang[0];
  prev_tang[1] = tang[1];
  prev_tang[2] = tang[2];

  /* and determine the new tangent vector */
  tang[0] = next_inter_pos[0] - inter_pos[0];
  tang[1] = next_inter_pos[1] - inter_pos[1];
  tang[2] = next_inter_pos[2] - inter_pos[2];  

  // norm the tangent-vector in relation to its projection into the x-z-plane
  tang_length = sqrtf(tang[0]*tang[0] + tang[2]*tang[2]);

  tang[0] /= tang_length;
  tang[1] /= tang_length;
  tang[2] /= tang_length;

  /* valid value */
  return GL_TRUE;
}


/****************************************************************************************/
/* do_display								                                            */
/****************************************************************************************/
/* renders the coaster track as a serie of quads along the given trajectory 			*/
/****************************************************************************************/
void do_display(void)
{
  /* draw the floor without texture */
  glDisable(GL_TEXTURE_2D);
  glBegin(GL_QUADS);
	glVertex3f((float)-2.0, (float)-0.2, (float)10.0);
	glVertex3f((float)132.0, (float)-0.2, (float)10.0);
	glVertex3f((float)132.0, (float)-0.2, (float)-142.0);
	glVertex3f((float)-2.0, (float)-0.2, (float)-142.0);
  glEnd();	
  glEnable(GL_TEXTURE_2D);

  /* init the trajectory for constructing the object trajectory */
  initTrajectoryValue(GEOMETRY_SUBDIVS);
  
  /* as long as there are valid trajectory positions ... */
  while(incrementTrajectoryValue())
  {
	/* ...draw quads between the first interpolated position (inter_pos) and the second follwing interpolated position (next_inter_pos) */
	/* and define the appropriate texture coordinates */
	glBegin(GL_QUADS); 

	  glTexCoord2f(0.0, 0.0); 
	  glVertex3f(inter_pos[0] - TRACK_BREADTH*prev_tang[2], inter_pos[1], inter_pos[2] + TRACK_BREADTH*prev_tang[0]);

	  glTexCoord2f(0.0, 1.0); 
	  glVertex3f(inter_pos[0] + TRACK_BREADTH*prev_tang[2], inter_pos[1], inter_pos[2] - TRACK_BREADTH*prev_tang[0]);

	  glTexCoord2f(1.0, 1.0); 
	  glVertex3f(next_inter_pos[0] + TRACK_BREADTH*tang[2], next_inter_pos[1], next_inter_pos[2] - TRACK_BREADTH*tang[0]);

	  glTexCoord2f(1.0, 0.0); 
	  glVertex3f(next_inter_pos[0] - TRACK_BREADTH*tang[2], next_inter_pos[1], next_inter_pos[2] + TRACK_BREADTH*tang[0]);

	glEnd();
  } 
}    


/****************************************************************************************/
/* initGraphics																		    */
/****************************************************************************************/
/* creates the graphics as a display list.										     	*/
/****************************************************************************************/
void initGraphics (void) {
	/* init the texturing */
    initTexturing();
    
    /* define the object as a precompiled display list */
    glNewList(1, GL_COMPILE);
		do_display();
	glEndList();
}


/****************************************************************************************/
/* setCamera								                                            */
/****************************************************************************************/
/* defines the new camera position and view direction                      				*/
/****************************************************************************************/
void setCamera(void)
{  
  /* if the animation of the camera is not in the stop state... */
  if(!stop)
  { 
	// ...then increment the trajectory position and check if it was the last value
	if(!incrementTrajectoryValue())
	{
	  initTrajectoryValue(TRAJECTORY_SUBDIVS);
	}
  }
  
  /* switch to the projection matrix */
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  
  // set the view frustum
  gluPerspective(80.0, WINDOW_WIDTH/WINDOW_HEIGHT, 0.1, 10000.0);

  // set the camera position and the view direction
  gluLookAt(inter_pos[0] , inter_pos[1]+CAMERA_HIGHT, inter_pos[2], 
			next_inter_pos[0], next_inter_pos[1]+CAMERA_HIGHT, next_inter_pos[2], 0.0, 1.0, 0.0);
}


/****************************************************************************************/
/* display: is the glut diplay fuction                                                  */
/****************************************************************************************/
/*          it sets the current camera position and displays the scene					*/
/****************************************************************************************/
void display(void)
{
  /* clear the window to draw in */
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  /* set the current camera position */
  //setCamera();

  /* draw the scene (precompiled in a display list */
  glCallList(1);

  /* cause the rendered image to be displayed */
  glFlush();
  glutSwapBuffers();
}

 
/****************************************************************************************/
/* animate: is the glut idle fuction                                                    */
/****************************************************************************************/
/*          it posts a redisplay.														*/
/*          it is executed when the cpu is idle.    							        */
/****************************************************************************************/
void animate(void)
{
  /* causes a repeated redisplay of the scene -> animation */
  glutPostRedisplay();
}


/****************************************************************************************/
/* timer: is the glut timer fuction                                                     */
/****************************************************************************************/
/*          this function is executed m miliseconds after its call.                     */
/*          it is used to trigger events like the update of a simulation.               */
/****************************************************************************************/
void timer(int m) {
	
	// reschedule the function
	glutTimerFunc(m,timer,m);
	
	// update camera position
	setCamera();
}


/****************************************************************************************/
/* reshape: is the glut reshape fuction                                                 */
/****************************************************************************************/
/*          it ajusts the viewing parameters after a reshaping of the window	     	*/
/****************************************************************************************/
/*  --> w   the new width of the window	     	                                        */
/*  --> h   the new height of the window	     	                                    */
/****************************************************************************************/
void reshape(int w, int h)
{
	/* set the new viewport dimensions */
    glViewport(0, 0, w, h);
    
    /* switch to the projection matrix mode and define the new projection matrix */
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(60.0, 1.0*(GLfloat)w/(GLfloat)h, 1.0, 30000.0);
    
    /* switch back to the model view matrix */
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
}


/****************************************************************************************/
/* specialKeys: is the glut special keyboard callback fuction                        */
/****************************************************************************************/
/*                 it determines which key was pressed and changes the corresponding	*/
/*                 parameters															*/
/****************************************************************************************/
/*       --> key   ascci value of the key which was pressed	     	                    */
/*       --> x, z  we don't care about them   	     	                                */
/****************************************************************************************/
void specialKeys(int key, int x, int z)
{
  /* determine the pressed key */    
  switch( key )
  {
	/* stop moving */
	case GLUT_KEY_UP: 
	{
	  stop = GL_FALSE;
	  return;
	}	  
	break;

	/* continue moving */
	case GLUT_KEY_DOWN: 
	{
	  stop = GL_TRUE;
	  return;
	}	  
	break;

	/* change texture */
	case GLUT_KEY_F1:
	  texture_type = TX_COLORED;
	break; 
	case GLUT_KEY_F2:
	  texture_type = TX_CHECKER_BOARD;
	break; 
	case GLUT_KEY_F3:
	  texture_type = TX_FILE;
	break;

	/* change filtering for mip map generation */
	case GLUT_KEY_F5:
	  txMipMapGenFilter = TX_NEAREST_NEIGHBOR_FILTER;
	break;   
	case GLUT_KEY_F6:
	  txMipMapGenFilter = TX_BOX_FILTER;
	break; 
	case GLUT_KEY_F7:
	  txMipMapGenFilter = TX_BILINEAR_FILTER;
	break;      	
  } 
  
  /* initialize the texturing */
  initTexturing();
}  



/****************************************************************************************/
/* keyboard: is the glut keyboard (non special keys) callback fuction                   */
/****************************************************************************************/
/*                 it determines which key was pressed and changes the corresponding	*/
/*                 parameters															*/
/****************************************************************************************/
/*       --> key   ascci value of the key which was pressed	     	                    */
/*       --> x, z  we don't care about them   	     	                                */
/****************************************************************************************/
void keyboard(unsigned char key, int x, int z)
{
  use_anisotropic = false;

  /* determine the pressed key */    
  switch( key )
  {
	
	/* change mip map filtering */
	case '1':
	  txMipMapFilter = GL_NEAREST;
	  use_anisotropic = false;
	break; 
	case '2':
	  txMipMapFilter = GL_NEAREST_MIPMAP_NEAREST;
	  use_anisotropic = false;
	break;   
	case '3':
	  txMipMapFilter = GL_LINEAR_MIPMAP_NEAREST;
	  use_anisotropic = false;
	break; 	
	case '4':
	  txMipMapFilter = GL_LINEAR_MIPMAP_LINEAR;
	  use_anisotropic = false;
	break;   
	case '5':
	  if (anisotropic_supported) use_anisotropic = true;
	break;   

	case 27:
	  exit(0);
	break;
  } 
  
  /* initialize the texturing */
  initTexturing();
}  
  	  
  	  
/****************************************************************************************/
/* main: is the main programm                                                           */
/****************************************************************************************/
/*       it initializes the glut callback functions, the display-list, as well as	    */
/*       the camera trajectory before entering the glut main loop			            */
/****************************************************************************************/
/*  --> argc, argv:  the console input is just passed to glutInit    	                */
/****************************************************************************************/
void main(int argc, char** argv)
{
	/* init the glut environment */
    glutInit(&argc, argv);
    
    /* init the display mode, double buffering is used for a smooth animation */
    glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
    
    /* init the glut window */
    glutInitWindowSize (WINDOW_WIDTH, WINDOW_HEIGHT);
    glutCreateWindow (argv[0]);
    
    initGraphics();

	/* register the reshape callback function */
    glutReshapeFunc (reshape);
    
    /* register the display callback function */
    glutDisplayFunc(display);
    
    /* register the idle function, which is executed whenever glut has nothing else to do */
    glutIdleFunc(animate);

	glutTimerFunc(camera_speed,timer,camera_speed);
    
    /* register the keyboard callback functions */
	glutKeyboardFunc(keyboard);
	glutSpecialFunc(specialKeys);  

	/* init the trajectory to use as camera path */
	initTrajectoryValue(TRAJECTORY_SUBDIVS);
	
	/* print the keyboard commands */
	printf("use the following keys:\n\n");
	printf("texture                   mipmap generation filter   mipmap filter\n");	
	printf("-------                   ------------------------   -------------\n\n");	
	printf("F1 = colored mipmap       F5 = nearest neighbor      1 = no\n");
	printf("F2 = checker board        F6 = box filter            2 = neaerest neighbor\n");
	printf("F3 = texture from file    F7 = bilinear filter       3 = bilinear\n");
	printf("                                                     4 = trilinear\n");		
	printf("                                                     5 = anisotropic\n");		
	printf("CURSOR_DOWN = stop        CURSOR_UP = continue\n\n");
	
	/* enter the glut main loop */
    glutMainLoop();  
}
